//  ComputationalAssembler.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;


/**
 * The ComputationalAssembler class instances make 
 * computational processes; we have a computationa assembler instance for
 * each unit
 *
 * @author Pietro Terna 
 */
public class ComputationalAssembler extends ComputationalAssemblerBasic{

    /**
     * the constructor for ComputationalAssembler
     */

    public ComputationalAssembler (Zone aZone, OrderDistiller od, ESFrameModelSwarm mo)
    {
	// Call the constructor for the parent class.
	super(aZone, od, mo);


    }

    //
    // this version of ComputationalAssembler.java is for tutorial-step3b
    //

    /** random position generation
     *  <br><br>
     *  the computational step 1201 (a code for the prey-predator tutorial) acts via 
     *  c1101 and creates a new random unit in the model stratum reported in position 
     *  0,1 of the first received matrix, with the probability set in position 0,0 of 
     *  the same matrix and increases the position 0,2 of the second received matrix 
     *  by 1, to count the created units;
     */
    public void c1201()
    {
	float count;
	Unit aUnit;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1201 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	// the number of matrixes in c1101 is 1, so we avoid the next check
	checkMatrixNumber=false;
	aUnit =	c1101();
	checkMatrixNumber=true;

	// if the unit has been created
	if(aUnit != null)
	    {
		mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
		    getMemoryMatrixAddress(1);
		count=mm1.getValue(layer,0+rd,2+cd);
		mm1.setValue(layer,0+rd,2+cd,++count);
	    }

    } // end c1201


    /** eating 
     *  <br><br>
     *  computational operations with code -1202 (a code for the prey-predator tutorial)
     *  <br><br>
     *
     *  this computational code acts via -1102 code to drop the unit the order is in, 
     *  with the probability set in position 0,0 of the first received matrix and 
     *  increases the  position 0,1 of the second received matrix by 1, to count the 
     *  dropped units; finally, it adds 1 to pos. 0,0 of the sending unit 
     *  memory matrix, to count its eating acts; done is set to true in c1102
     */
    public void c1202()
    {
	Order theCurrentOrder;
	Unit theUnitTheOrderComesFrom;
	int layer=0;
	float count;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1202 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	c1102();

	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
	theUnitTheOrderComesFrom=theCurrentOrder.getTheUnitTheOrderComesFrom();
	count=(theUnitTheOrderComesFrom.getMemoryMatrix()).getValue(layer,0,0);
	count++;
	(theUnitTheOrderComesFrom.getMemoryMatrix()).setValue(layer,0,0,count);

    } // end c1202


    /** perishing (unsufficient food, i.e. energy)
     *  <br><br>
     *  computational operations with code -1204 (a code for the prey-predator tutorial)
     *  <br><br>
     *
     *  this computational code acts via c1102 code to drop the unit the order is in, if
     *  pos. 0,0 value of the unit  memory matrix is less then the parameter in 1,2 
     *  (minEnergy) of the second received matrix; c1102 acts with the probability set 
     *  in position 1,0 of the first received matrix (internal trick: c1102 looks for 
     *  0+rd,0+cd using temporary rd (the row displacement) set to 1; c1102 increases 
     *  the  position 1,1 (always via the displacement, being the internal coordinates 
     *  0+rd,1+cd, of the second received matrix by 1, to count the dropped units; 
     *  done is set to true in c1102 or here in c1204
     */
    public void c1204()
    {
	Order theCurrentOrder;
	Unit theUnitTheOrderIsIn;
	int layer=0; float minEnergy;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1204 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);

	minEnergy=mm0.getValue(layer,1,2);

	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
	theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();

	if (minEnergy > (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,0,0))
	    {
		rd=1; // prob is in 1,0, but c1102() looks for it at 0+rd,0+cd (with cd set to 0)
		      // also the count of the dropped units is in 0+rd,1+cd (with cd set to 0)
		c1102();
		rd=0;
	    }
	else done=true; // doing nothing

    } // end c1204

    /** consuming (decreasing energy)
     *  <br><br>
     *  computational operations with code -1205 (a code for the prey-predator tutorial)
     *  <br><br>
     *
     *  this computational code acts with the probability set in position 2,0 of the unique
     *  received matrix to decrease of 1 unit the position 0,0 of the memory matrix of the unit
     *  the order is in; 
     *  finally done is set to true
     */
    public void c1205()
    {
	Order theCurrentOrder;
	Unit theUnitTheOrderIsIn;
	int layer=0; float prob, energy;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1)
	    {
		System.out.println("Code -1205 requires 1 matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	prob = mm0.getValue(layer,2,0);

	if(prob >= Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    {
		theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
		theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
		// getting unit energy and decreasing it
		energy=(theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,0,0);
		(theUnitTheOrderIsIn.getMemoryMatrix()).setValue(layer,0,0,energy-1);
	    }

	done=true;

    } // end c1205


    /** reproducing (creating a copy of themselves near to their position, if their 
	energy is sufficient)
     *  <br><br>
     *  computational operations with code -1206 (a code for the prey-predator tutorial)
     *  <br><br>
     *
     *  this computational code acts via the c1103 code to create, with the probability set in 
     *  position 3,0 of the first received matrix (c1103 see it with rd=3 as 0+rd,0+cd), a copy
     *  of the unit the order is in, near to it, if comparing position 0,0 of the unit memory 
     *  matrix and position 3,2 of the second received matrix, i.e. if energy >= requested level 
     *  to reproduce; the count of the created unit is reported in position 3,1 of the second
     *  received matrix  (via c1103 with rd=3)
     *
     *  done is set to true in c1103 or here in c1206
     */
    public void c1206()
    {
	Order theCurrentOrder;
	Unit theUnitTheOrderIsIn;
	int layer=0; float energyReproducingLevel, energy;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1206 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	energyReproducingLevel = mm0.getValue(layer,3,2);

	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
	theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
	// getting unit energy and decreasing it
	energy=(theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,0,0);
	
	// reproducing if energy is > the energyReproducingLevel
	if (energy>=energyReproducingLevel)
	    {
		rd=3;
		c1103();
		rd=0;
	    }
	else done=true;

    } // end c1206

    /** 
     *  the computational code c1207 (a code for the prey-predator tutorial) uses c1206, 
     *  but acting only if the unit the order is in, is in a specific stratum (that 
     *  reported in position 0,3 of the second received matrix)
     */
    public void c1207()
    {
	Unit theUnitTheOrderIsIn;
	Order theCurrentOrder;
	int ourStratum;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1207 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	// is the unit in the correct stratum?
	mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	ourStratum= (int) mm1.getValue(layer,0+3,3);

	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
	theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();

	if(ourStratum==((ESFrameModelSwarm)theUnitTheOrderIsIn.
			getUnitModel()).getStratum()) {
	    c1206();
	}
	else done=true;

    } // end c1207

    /** growing (creating a copy of themselves in the same position, in an other model 
     *  stratum, if their energy is sufficient)
     *  <br><br>
     *  computational operations with code -1210 (a code for the prey-predator tutorial)
     *  <br><br>
     *
     *  this computational step acts via c1110 and creates a new unit in the same x,y 
     *  position, but in the model stratum reported in position 0,1 of the first received 
     *  matrix, with the probability set in position 0,0 of the same matrix and increases 
     *  the position 0,2 of the second received matrix by 1, to count the created units; 
     *
     *  c1110, called by c1210, acts if a specific cell in the unit memory matrix is >0; 
     *  this cell is identified via the urd, ucd trick (the same used for rd and cd, but 
     *  applied to the unit memory matrix) in c1210, where also its content is calculated;
     *  the specific cell content is set > 0 if the value (energy) in 0,0 of unit memory 
     *  matrix is > the level set in position 0,3 ot the second received matrix
     *
     *  the done assignment is in c1110 or here; 
     *
     */
    public void c1210()
    {
	Unit theUnitTheOrderIsIn;
	Order theCurrentOrder;
	MemoryMatrix mU;
	int layer=0; 
	float count, limit;
	Unit aUnit;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1210 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	// the growing unit 
	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
	theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();

	// second received matrix
	mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	rd=4;
	limit=mm1.getValue(layer,0+rd,3+cd);

	// growing if the unit energy is > the limit value
	mU=theUnitTheOrderIsIn.getMemoryMatrix();
	if(mU.getValue(layer,0,0)>limit) mU.setValue(layer,0,1,1.0); // 1.0 as a generic positive
	else
	    {
		done=true;
		return;
	    }


	checkMatrixNumber=false;
	ucd=1; // check pos 0,1 in unit memory matrix
	aUnit=c1110();
	checkMatrixNumber=true;
	ucd=0;

	// if the unit has been created
	if(aUnit != null)
	    {
		count=mm1.getValue(layer,0+rd,2+cd);
		mm1.setValue(layer,0+rd,2+cd,++count);
	    }
	rd=0;


    } // end c1210


    /** this code simply call the highly usable c1199 computational code, using rd=1 to direct
     *  the called step to the correct row of the input matrixes 
     */
    public void c1299(){

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1299 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	rd=1;
	c1199();
	rd=0;

    } // end c1299

}
